"""
define data structures to use in radnlp
"""
from collections import namedtuple

classrslts = namedtuple("classrslts",
                        ["context_document",
                         "exam_type",
                         "report_text",
                         "classification_result"])
def get_brief_classification(crslt):
    """
    crslt: dictionary of classification results
    """
    return {cat: cla[0] for cat, cla in crslt.items()}

def filter_classification_rslts(crslt, select = (7,8)):
    """
    crslt: dictionary of classifications
    select: numeric classification values to keep
    """
    return {cat: cla for cat, cla in crslt.items() if cla[0] in select}

    
def get_classification_result_brief(crslt):
    """
    crslt: an instance of classrslts

    returns a string with each category/classification pair
    """
    return {cat: cla[0] for cat,cla in
                      crslt.classification_result.items()}



import gzip
import json
try:
    import pickle
except:
    import cPickle as pickle


def to_pickle(data, fname, compress=True):
    """
    save analysis to pickle file
    """
    if compress:
        with gzip.open(fname, "wb") as f0:
            pickle.dump(data, f0)
    else:
        with open(fname, "wb") as f0:
            pickle.dump(data, f0)


def to_json(data, fname, compress=True):
    """
    save analysis to json file
    """
    if compress:
        with gzip.open(fname, "wb") as f0:
            json.dump(data, f0)
    else:
        with open(fname, "wb") as f0:
            json.dump(data, f0)


def read_pickle(fname, compress=True):
    """
    read analysis from pickle file
    """
    if compress:
        with gzip.open(fname, "rb") as f0:
            return pickle.load(f0)
    else:
        with open(fname, "rb") as f0:
            return pickle.load(f0)


def read_json(data, fname, compress=True):
    """
    save analysis to json file
    """
    if compress:
        with gzip.open(fname, "wb") as f0:
            json.load(f0)
    else:
        with open(fname, "wb") as f0:
            json.load(f0)



"""
utilities to edit pyConTextNLP itemData and other stuff in the
Jupyter notebook
"""
from IPython.display import display, HTML, clear_output
import pandas as pd

def get_multi_attribute(i):

    cat = input("Enter %s:\n"%i)
    if isinstance(cat,str):
        return cat
    else:
        return ",".join(cat)
def get_attribute(att):
    attribute = input("Enter value for %s:\n"%att)
    return attribute
def should_edit_attribute(i, attr):
    """
    i: a pandas series containing an individual item
    attr: the attribute to consider editing

    prompt user whether the named attribute should be edited
    """
    clear_output()
    print(i)
    yn = input("Edit %s?\n"%attr)
    try:
        return bool(yn.lower()[0] == 'y')
    except IndexError:
        return edit_attribute(attr)

def create_item():
    """
    create an itemData tuple as a pandas series
    """
    lex = get_attribute("Lex")
    t = get_multi_attribute("Type")
    regex = get_attribute("Regex")
    direction = get_attribute("Direction")
    codes = get_multi_attribute("Codes")
    newitem=pd.Series(data=[lex, t, regex, direction, codes],
                      index=["Lex","Type","Regex","Direction","Codes"])
    return newitem


def edit_item(items, literal):
    """
    items: a pandas DataFrame containing itemData
    literal: a string to match lexical type
    """
    for index, _ in items[items["Lex"] == literal].iterrows():
        if should_edit_attribute(items.loc[index], "Lex"):
            print("editing lex")
            items.loc[index, "Lex"] = get_attribute("Lex")
        if should_edit_attribute(items.loc[index], "Type"):
            items.loc[index, "Type"] = get_multi_attribute("Type")
        if should_edit_attribute(items.loc[index], "Regex"):
            items.loc[index, "Regex"] = get_attribute("Regex")
        if should_edit_attribute(items.loc[index], "Direction"):
            items.loc[index, "Direction"] = get_attribute("Direction")
        if should_edit_attribute(items.loc[index], "Codes"):
            items.loc[index, "Codes"] = get_multi_attribute("Direction")



from . import utils
from . import schema
from . import split
from . import classify

def process_report(report, modifier, target,
                   classification_rules,
                   category_rules,
                   _schema,
                   **kwargs):
    """
    """
    markup = utils.mark_report(split.get_sentences(report),
                         modifier,
                         target)
    return  classify.classify_document_targets(markup,
                                          classification_rules,
                                          category_rules,
                                          severity_rules,
                                          _schema,
                                          kwargs)



import pyConTextNLP.pyConTextGraph as pyConText
import networkx as nx

def modifies(g, n, modifiers):
    """
    Tests whether any of the modifiers of node n are in
    any of the categories listed in 'modifiers'
    """
    pred = g.predecessors(n)
    if not pred:
        return False
    pcats = []
    for p in pred:
        pcats.extend(p.getCategory())
    return bool(set(pcats).intersection([m.lower() for m in modifiers]))


def matched_modifiers(g, n, modifiers):
    """
    returns the set of predecessors of node 'n' that are of a
    category contained in 'modifiers'
    """
    pred = g.predecessors(n)
    if not pred:
        return False
    pcats = []
    for p in pred:
        pcats.extend(p.getCategory())

    return set(pcats).intersection([m.lower() for m in modifiers])


def return_matched_modifiers(g, n, modifiers):
    """
    """
    pred = g.predecessors(n)
    if not pred:
        return []
    mods = [m.lower() for m in modifiers]
    mmods = [p for p in pred if p.isA(mods)]
    return mmods


def markup_sentence(s, modifiers, targets):
    """
    """
    markup = pyConText.ConTextMarkup()
    markup.setRawText(s)
    markup.cleanText()
    markup.markItems(modifiers, mode="modifier")
    markup.markItems(targets, mode="target")
    markup.pruneMarks()
    markup.dropMarks('Exclusion')
    # apply modifiers to any targets within the modifiers scope
    markup.applyModifiers()
    markup.pruneSelfModifyingRelationships()
    markup.dropInactiveModifiers()
    return markup

def _merge_markups(rmarkups):
    if rmarkups:
        raw_text = " ".join([r.graph["__rawTxt"] for r in rmarkups])
        text = " ".join([r.graph["__txt"] for r in rmarkups])
        newgraph = nx.compose_all(rmarkups)
        newgraph.graph["__rawTxt"] = raw_text
        newgraph.graph["__txt"] = text
        return newgraph
    else:
        return nx.DiGraph()
def mark_report(report, modifiers, targets):
    context = pyConText.ConTextDocument()
    for m in [markup_sentence(s,
                              modifiers,
                              targets) for s in report]:
        context.addMarkup(m)
    return context
    # return _merge_markups([markup_sentence(s,
    #                                       modifiers,
    #                                       targets) for s in report])


def get_severity(g, t, severity_rule):
    """
    """
    if not t.isA(severity_rule[0]):
        return []
    smods = return_matched_modifiers(g, t, severity_rule[1])
    if smods:
        severity_results = []
        for m in smods:
            mgd = m.getMatchedGroupDictionary()
            val = mgd.get('value')
            units = mgd.get('unit')
            phrase = m.getPhrase()
        severity_results.append((phrase, val, units))
        return severity_results
    else:
        return []


def anatomy_recategorize(g, t, category_rule):
    """
    create a new category based on category_rule
    g: a graph containing a markup
    t: a target (???)
    category_rule: the rule to apply for the recateogrization

    """
    if not t.isA(category_rule[0]):
        return

    mods = g.predecessors(t)

    if mods:
        mmods = matched_modifiers(g, t, category_rule[1])
        if mmods:
            new_category = []
            for m in mmods:
                nc = "_".join((m.lower(), category_rule[0]))
                new_category.append(nc)
            t.replaceCategory(category_rule[0], new_category)


def generic_classifier(g, t, rule):
    """
    based on the modifiers of the target 't' and the provide rule
    in 'rule' classify the target node
    """
    mods = g.predecessors(t)
    if not mods:
        return rule["DEFAULT"]
    for r in rule["RULES"]:
        if modifies(g, t, r[1]):
            return r[0]
    return rule["DEFAULT"]

def process_report(report, modifier, target,
                   classification_rules,
                   category_rules,
                   _schema,
                   **kwargs):
    """
    """
    markup = mark_report(split.get_sentences(report),
                         modifier,
                         target)
    return  classify_document_targets(markup,
                                          classification_rules,
                                          category_rules,
                                          severity_rules,
                                          _schema,
                                          kwargs)



"""
Tools for splitting report and recognizing sections
"""
import re
from textblob import TextBlob
from collections import OrderedDict

r_headings = re.compile(r"""(?P<heading>([A-Z ]+\s)?[A-Z()]+:)""")
r_digits = re.compile(r"""\d\.""")
r_enumerate = re.compile(r"""((\d(.|:|\))\s)(.+)(?=\n))""")

canned_phrases = ("if you have any question about this report, contact me at:",
                  "i have personally reviewed the images and agree",
                  "i have personally reviewed the images for this examination",
                  "please utilize the following guidance for the management of these patients",
                  "please utilize the following guidance for management",
                  "i have reviewed the images and approve this final report",
                 "the imaging exam has been reviewed and the report has been issued by a radiologist physician",
                 )
def terminate_lists(txt):
    """
    Replace enumerated lists that don't end with a period/question mark
    """
    lists = r_enumerate.findall(txt)
    for l in lists:
        txt = txt.replace(l[0],"%s."%l[0])
    return txt


def get_headings(txt):
    """
    """
    global r_headings
    return [r.group("heading").strip() for r in r_headings.finditer(txt)]
def find_terminating_sentences(txt,phrases=None):
    """

    """
    try:
        if not phrases:
            phrases = ()
        indices = [i for i in [txt.lower().find(p) for p in phrases] if i != -1]
        return min(indices)
    except:
        return None


def get_report(txt,terminating_phrases= None):
    """
    get report text up to any terminating_phrases
    """
    if not terminating_phrases:
        terminating_phrases = canned_phrases
    ti = find_terminating_sentences(txt,terminating_phrases)
    global r_digits
    if ti != -1:
        return r_digits.sub("",txt[:ti])
    else:
        return r_digits.sub("",txt)


def preprocess_report(txt,
                      fix_enumerated_lists=True,
                      drop_boiler_plate=True,
                      canned_phrases=None):
    """
    from a string (txt) containing a report, return the relevant
    portions

    fix_enumerated_lists: boolean.
        If true, use regular expressions to transform enumerated lists into more sentence like structures.
    drop_boiler_plate: boolean.
        If true, only return text preceding canned_phrases
    canned_phrases: None or list.
        List of canned phrases to exlcude as boilerplate
    """
    if fix_enumerated_lists:
        txt = terminate_lists(txt)
    if drop_boiler_plate:
        txt = get_report(txt,terminating_phrases=canned_phrases)
    return txt

def grab_impression(txt):
    """
    grab impression from text via looking for IMPRESSION:
    """
    try:
        return txt[txt.index("IMPRESSION:"):].split("IMPRESSION:")[1]
    except Exception as error:
        print(error)
        return ""


def  get_sections(txt, headings):
    h = headings.pop()


def get_section(txt,heading):
    try:
        loc = txt.index(heading)
        return txt[:loc], txt[loc:].split(heading)[1].strip()
    except Exception as error:
        #print(error, heading, txt)
        return txt, ""


def get_sections_by_headings(txt, sections, headings):
    if not headings:
        return txt, sections
    else:
        h = headings.pop()
        txt, sections[h] = get_section(txt, h)
        return get_sections_by_headings(txt, sections, headings)


def reverse_sections(secs):
    _secs = OrderedDict()
    while secs:
        sec,txt = secs.popitem()
        _secs[sec] = txt
    return _secs


def split_report(txt):
    headings = get_headings(txt)
    txt, secs = get_sections_by_headings(txt, OrderedDict(), headings)
    return reverse_sections(secs)

def get_sentences(report):
    """
    Wrapper around TextBlob
    generates a TextBlob instance from report and
    returns the sentences
    """
    return [s.raw for s in TextBlob(report).sentences]



"""
Functions and variables for viewing marked up reports
"""
from pyConTextNLP.display.html import mark_document_with_html
from . data import get_classification_result_brief as gcrb
import networkx as nx
from . data import get_brief_classification
clrs = {\
"finding": "blue",
"definite_negated_existence": "red",
"probable_negated_existence": "indianred",
"ambivalent_existence": "orange",
"probable_existence": "forestgreen",
"definite_existence": "green",
"historical": "goldenrod",
"acute": "golden"
}

codingKey = {\
1: "AMBIVALENT",
2: "Negative/Certain/Acute",
3: "Negative/Uncertain/Chronic",
4: "Positive/Uncertain/Chronic",
5: "Positive/Certain/Chronic",
6: "Negative/Uncertain/Acute",
7: "Positive/Uncertain/Acute",
8: "Positive/Certain/Acute"
}

def classrslt_to_str(crslt, joiner="\n"):
    """

    """

    class_rslts = get_brief_classification(crslt)
    class_str = joiner.join(["%s: %s"%(k,codingKey.get(v,"NA")) for
                           k, v in class_rslts.items()])

    return class_str
def classrslt_to_html(crslt, joiner="<br><br>"):
    """

    """

    class_rslts = get_brief_classification(crslt)
    class_str = joiner.join(["<b>%s</b> (%s)"%(k,codingKey.get(v,"NA")) for
                           k, v in class_rslts.items()])

    return class_str
def classification_to_html(markup_rslt, joiner="<br><br>"):
    """

    """

    class_rslts = gcrb(markup_rslt)
    class_str = joiner.join(["<b>%s</b> (%s)"%(k,codingKey.get(v,"NA")) for
                           k, v in class_rslts.items()])

    return class_str
def markup_to_html(markup_rslt, color_map=None, joiner="<br><br>"):
    """

    """
    if not color_map:
        color_map = clrs

    mt = mark_document_with_html(markup_rslt.context_document,
                                 colors=color_map)
    class_rslts = gcrb(markup_rslt)
    class_str = joiner.join(["<b>%s</b> (%s)"%(k,codingKey.get(v,"NA")) for
                           k, v in class_rslts.items()])
    txt = """
          <table style="width:100">
          <caption>PE Finder Case Review</caption>
          <tr><th>report</th><th>classification</th></tr>
          <tr><td width='500'>%s</td><td width='400'>%s</td></tr>
          </table>""" % (mt,class_str)
    return txt

def get_node_brief(n,id_digits=3):
    return """%s %s(%s)"""%(n.getPhrase(),
                             n.getCategory(),
                             ("%d"%n.getTagID())[-id_digits:])

def documentgraph_to_viewgraph(dg):
    newg = nx.DiGraph()
    edges = dg.edges()
    for e in edges:
        newg.add_edge(get_node_brief(e[0]),
                      get_node_brief(e[1]))
    return newg
def get_pydot_graph(markup_rslt):

    return nx.nx_pydot.to_pydot(
        documentgraph_to_viewgraph(
            markup_rslt.context_document.getDocumentGraph())
        )


def markup_to_pydot(markup_rslt,fname="./tmp.png"):
    ag = get_pydot_graph(markup_rslt)
    ag.write_png(fname)
    return fname



"""This module containes functions for reading rules and schema for document
classification based on pyConTextNLP parsing"""
import platform

if platform.python_version_tuple()[0] == '2':
    import urllib2.urlopen as urlopen
    import urllib2.urlparse as urlparse
else:
    from urllib.request import urlopen
    import urllib.parse as urlparse

def read_rules(fname):
    """read the sentence level rules"""

    p = urlparse.urlparse(fname)

    if not p.scheme:
        fname = "file://"+fname

    f0 = urlopen(fname)

    data = f0.readlines()
    class_rules = {}
    category_rules = []
    severity_rules = []
    for d in data:
        if platform.python_version_tuple()[0] == '2':
            tmp = d.strip().split(",")
        else:
            tmp = d.decode('utf-8').strip().split(",")

        if not tmp[0][0] == "#":  # '#' comment character
            if tmp[0] == "@CLASSIFICATION_RULE":
                r = class_rules.get(tmp[1],
                                    {"LABEL":"",
                                     "DEFAULT":"",
                                     "RULES":[]})
                value = int(tmp[3])
                if tmp[2] == 'DEFAULT':
                    r["DEFAULT"] = value
                elif tmp[2] == 'RULE':
                    rcs = []
                    for rc in tmp[4:]:
                        rcs.append(rc)
                    r["RULES"].append((value, rcs))
                class_rules[tmp[1]] = r
            elif tmp[0] == "@CATEGORY_RULE":
                category_rules.append((tmp[1], [r for r in tmp[2:]]))
            elif tmp[0] == "@SEVERITY_RULE":
                severity_rules.append((tmp[1], [r for r in tmp[2:]]))

    return class_rules, category_rules, severity_rules



"""
This module implements a generic document classification based on
pyConTextNLP markup
"""

from . import utils
from . import schema


"""
def modify_categories(document, category_rules):
    for cr in category_rules:
        utils.anatomy_recategorize(g,t,cr)
"""


def classify_result(doc_rslts, _schema):
        """
        given results in doc_rslts compare to classification_schema and
        return score.
        Takes a three-tuple of boolean values for
            * Disease State Positive
            * Disease State Certain
            * Disease State Acute
        """
        return schema.assign_schema(doc_rslts, _schema)


def classify_document_targets(doc,
                              classification_rules,
                              category_rules,
                              severity_rules,
                              _schema,
                              neg_filters=["definite_negated_existence",
                                           "probable_negated_existence"],
                              exclusion_categories=["QUALITY_FEATURE",
                                                    "ARTIFACT"]):
    """
    Look at the targets and their modifiers to get an overall
    classification for the document_markup
    """
    rslts = {}

    qualityInducedUncertainty = False
    g = doc.getDocumentGraph()
    
    targets = [n[0] for n in g.nodes(data=True)
               if n[1].get("category", "") == 'target']

    if targets:
        for t in targets:
            severity_values = []
            current_rslts = {}
            current_category = t.getCategory()
            if not t.isA(exclusion_categories):
                # iterate across all the classification_rules
                current_rslts = \
                    {rk:utils.generic_classifier(g,
                                                 t,
                                                 classification_rules[rk])
                     for rk in classification_rules}

                for cr in category_rules:
                    utils.anatomy_recategorize(g, t, cr)
                for sv in severity_rules:
                    severity = utils.get_severity(g, t, sv)
                    severity_values.extend(severity)
                current_category = t.categoryString()
                # now need to compare current_rslts to rslts
                # to select most Positive
                docr = classify_result(current_rslts, _schema)
                trslts = rslts.get(current_category, (-1, '', []))
                if trslts[0] < docr:
                    trslts = (docr, t.getXML(), severity_values)
                rslts[current_category] = trslts
            else:
                if t.isA('QUALITY_FEATURE'):
                    qualityInducedUncertainty = True
                else:
                    # if non-negated artifacts call uncertain
                    if not utils.modifies(g, t, neg_filters):
                        qualityInducedUncertainty = True
    return rslts



"""
Functions for reading and applying schema for document classification
"""
import platform

if platform.python_version_tuple()[0] == '2':
    import urllib2.urlopen as urlopen
    import urllib2.urlparse as urlparse
else:
    from urllib.request import urlopen
    import urllib.parse as urlparse


def instantiate_schema(values, rule):
    """
    evaluates rule by substituting values into rule and evaluating the resulting literal.
    This is currently insecure
        * "For security the ast.literal_eval() method should be used."
    """
    r = rule
    for k in values.keys():
        r = r.replace(k, values[k].__str__())
    #return ast.literal_eval(r)
    return eval(r)

def assign_schema(values, rules):
    """
    """
    for k in rules.keys():
        if instantiate_schema(values, rules[k][1]):
            return k
def read_schema(fname):
    """
    read the schema and schema classifier functionality
    """

    p = urlparse.urlparse(fname)
    if not p.scheme:
        fname = "file://"+fname

    f0 = urlopen(fname)
    schema = {}
    tmp = f0.readlines()
    for t in tmp:
        if platform.python_version_tuple()[0] == '2':
            t = t.strip().split(",")
        else:
            t = t.decode('utf-8').strip().split(",")
        if t[0][0] != "#":
            schema[int(t[0])] = (t[1], t[2])
    return schema
